import { createApp, reactive, readonly } from "vue";
import Resolver from "./Resolver.vue";
import sharedData from "./sharedData.js";

const windowName = new URLSearchParams(__resourceQuery).get("window");
const backgroundName = new URLSearchParams(__resourceQuery).get("background");

const isBackgroundProcess = backgroundName === windowName;

window.windowName = windowName;

/**
 * due to limitations of Proxy and vue.js 3 we need to create a
 * window.sharedData object in each window and make sure every time the object
 * inside the background process window is updated, the other objects are synced
 * with it. While this does lead to data duplication across window of sharedData,
 * the memory impact (depending on use of sharedData) should be low.
 *
 * The limitation of this solution is shared data should be modified only
 * in the background window under this.$state while in other windows it will be
 * readonly, thus all mutations on $state should always happen in background.
**/

if (window.overwolf) {
	// Initialize _sharedData with imported data from sharedData
	window._sharedData = sharedData;

	if (isBackgroundProcess) {
		// Initialize sharedData with _sharedData in the background process window (the first starting window)
		window.sharedData = window._sharedData;
	} else {
		// Every time a new window starts, we need to link the sharedData from background to this new window
		const coreWindow = window.overwolf.windows.getMainWindow();

		// first, we initialize the sharedData in the new window with a reactive/proxified
		// clone of the current value of _sharedData from the background window
		// thus creating an initial state
		window.sharedData = reactive(coreWindow._sharedData);

		// then we redefine the background window's sharedData as a Proxy wrapper
		// on top of it that interccepts all set's (modifications) and propagates
		// them to this new window by updating the new window's sharedData too
		// using overwolf's windows.getOpenWindows for the given window
		coreWindow.sharedData = new Proxy(
			coreWindow.sharedData,
			{
				set(obj, prop, value) {
					console.log("CHANGE DETECTED"); // Make sure this is printed the same number of times in all desktop windows

					overwolf.windows.getOpenWindows(windows => {
						// make sure the window exists, in case it was closed to avoid errors
						if (windows[windowName] && windows[windowName].sharedData) {
							windows[windowName].sharedData[prop] = value;
						}
					});

					return Reflect.set(...arguments);
				}
			}
		);

		// we also have to relink the this.$state to the new Proxified sharedData
		coreWindow.App.config.globalProperties.$state = coreWindow.sharedData;
	}
} else {
	window.sharedData = sharedData;
}

const app = createApp(Resolver);

// expose the sharedData as $state in each window
if (window.overwolf) {
	if (isBackgroundProcess) {
		// directly in background window
		app.config.globalProperties.$state = window.sharedData;
	} else {
		// as readonly in all other windows
		app.config.globalProperties.$state = readonly(window.sharedData);
	}
}

if (window.overwolf) {
	app.mount("#app");
	window.App = app; // expose vue app as window.App so we can update $state
} else {
	// run window in chrome
	import("@app/windows/background.vue").then(module => {
		const bgd = createApp(module.default)
		bgd.config.globalProperties.$state = reactive(window.sharedData)
		bgd.mount('#app')
		app.config.globalProperties.$state = readonly(reactive(window.sharedData))
		app.mount('#app')
		window.App = app
	})
}
